import pytest

from a3m.client.mcp import handle_batch_task


@pytest.mark.django_db
def test_handle_batch_task_replaces_non_ascii_arguments(mocker):
    # We are only interested in verifying the string replacement logic
    # for task arguments and mock the remaining functionality
    mocker.patch("a3m.client.mcp.Job")
    mocker.patch("a3m.client.mcp.Task")
    mocker.patch("a3m.client.mcp.retryOnFailure")

    # This is the only function that uses the arguments after the replacements
    _parse_command_line = mocker.patch("a3m.client.mcp._parse_command_line")

    # The mocked module will not have a `concurrent_instances` attribute
    mocker.patch(
        "importlib.import_module", return_value=mocker.MagicMock(spec=["call"])
    )

    # Mock the two parameters sent to handle_batch_task
    task_name = "tásk".encode()
    batch_payload = {
        "tasks": {
            "some_task_uuid": {
                "uuid": "some_task_uuid",
                "arguments": "montréal %taskUUID% %jobCreatedDate%",
                "createdDate": "some montréal datetime",
                "wants_output": False,
                "execute": "command",
            }
        }
    }
    handle_batch_task(task_name, batch_payload)

    # Check that string replacement were successful
    _parse_command_line.assert_called_once_with(
        "montréal some_task_uuid some montréal datetime"
    )



import os
import re
import sys
from datetime import datetime

sys.path.append(os.path.abspath("../"))

from a3m.cli.common import init_django

init_django()

needs_sphinx = "3.2"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinxcontrib.mermaid",
]

autoclass_content = "both"
autodoc_member_order = "bysource"
source_suffix = ".rst"
master_doc = "index"
project = "a3m"
author = "%d Artefactual Systems Inc." % datetime.now().year

output = os.popen("git describe --tags --abbrev=0").read().strip()  # nosec
release = re.sub("^v", "", output)
version = release

language = "en"
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
pygments_style = "sphinx"
todo_include_todos = True

html_theme = "alabaster"
html_theme_options = {
    "description": "Lightweight Archivematica",
    "fixed_sidebar": True,
    "github_user": "artefactual-labs",
    "github_repo": "a3m",
    "github_banner": False,
    "github_button": False,
}
html_static_path = ["_static"]
htmlhelp_basename = "a3mdoc"

suppress_warnings = ["image.nonlocal_uri"]

mermaid_version = "8.8.2"



import pathlib

import pytest


@pytest.fixture(autouse=True)
def set_xml_catalog_files(monkeypatch):
    """Use local XML schemas for validation."""
    monkeypatch.setenv(
        "XML_CATALOG_FILES",
        str(
            pathlib.Path(__file__).parent.parent
            / "a3m/client/assets/catalog/catalog.xml"
        ),
    )



import uuid

import pytest

from a3m.client.clientScripts.identify_file_format import main
from a3m.main.models import Event
from a3m.main.models import File
from a3m.main.models import FileID
from a3m.main.models import Transfer


@pytest.fixture()
def subdir_path(tmp_path):
    subdir = tmp_path / "subdir1"
    subdir.mkdir()

    return subdir


@pytest.fixture()
def file_path(subdir_path):
    file_path = subdir_path / "script.py"
    file_path.write_text("import sys; sys.exit(0)")

    return file_path


@pytest.fixture()
def transfer(db):
    return Transfer.objects.create(
        uuid=uuid.uuid4(), currentlocation=r"%transferDirectory%"
    )


@pytest.fixture()
def file_obj(db, transfer, tmp_path, file_path):
    file_obj_path = "".join(
        [transfer.currentlocation, str(file_path.relative_to(tmp_path))]
    )
    file_obj = File.objects.create(
        uuid=uuid.uuid4(),
        transfer=transfer,
        originallocation=file_obj_path,
        currentlocation=file_obj_path,
        removedtime=None,
        size=24,
        checksum="7e272e3e7076fef4248bc278918ecf003f05f275dff3bdb9140f1f4120b76ff1",
        checksumtype="sha256",
    )

    return file_obj


def test_identify_file_format(file_obj, file_path):
    code = main(str(file_path), file_obj.uuid, disable_reidentify=False)
    assert code == 0

    format_version = file_obj.get_format_version()
    assert format_version is not None
    assert format_version.pronom_id == "fmt/938"

    Event.objects.get(
        file_uuid=file_obj,
        event_type="format identification",
        event_outcome="Positive",
        event_outcome_detail="fmt/938",
    )
    FileID.objects.get(
        file_id=file_obj.uuid,
        format_name="Python Script File",
        format_registry_key="fmt/938",
    )



import concurrent.futures
import os
import threading
import uuid
from pathlib import Path

import pytest

from a3m.api.transferservice.v1beta1.request_response_pb2 import ProcessingConfig
from a3m.main import models
from a3m.server.packages import Package
from a3m.server.queues import PackageQueue
from a3m.server.workflow import load as load_workflow

FIXTURES_DIR = os.path.join(os.path.dirname(__file__), "fixtures")
INTEGRATION_TEST_PATH = os.path.join(FIXTURES_DIR, "workflow-integration-test.json")


@pytest.fixture
def workflow(request):
    with open(INTEGRATION_TEST_PATH) as workflow_file:
        return load_workflow(workflow_file)


@pytest.fixture
def package_queue(request):
    executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
    return PackageQueue(executor, threading.Event(), debug=True)


@pytest.fixture
def package(request, db, package_queue, workflow):
    return Package.create_package(
        package_queue,
        package_queue.executor,
        workflow,
        "name",
        "file:///tmp/foobar.gz",
        ProcessingConfig(),
    )


@pytest.mark.django_db(transaction=True)
def test_reload_file_list(tmp_path, package):
    path = tmp_path / f"test-transfer-{package.subid}"
    package.current_path = str(path)

    # Add files to the transfer to simulate a transfer existing on disk.
    path.mkdir()
    objects_path = path / "objects"
    objects_path.mkdir()
    first_file = objects_path / "file.txt"
    first_file.touch()

    kwargs = {
        "uuid": uuid.uuid4(),
        "currentlocation": "".join(
            [package.replacement_path_string, str(Path("/objects/file.txt"))]
        ),
        "filegrpuse": "original",
        "transfer_id": package.subid,
    }
    models.File.objects.create(**kwargs)
    assert models.File.objects.filter(transfer_id=str(package.subid)).count() == 1

    # Add a new file to the file-system, e.g. to simulate normalization for
    # preservation adding a new object.
    new_file = objects_path / "new_file.txt"
    new_file.touch()

    # One file will be returned from the database  with a UUID, another from
    # the filesystem without a UUID.
    for file_count, file_info in enumerate(package.files("/objects"), 1):
        assert "%fileUUID%" in file_info
        assert "%fileGrpUse%" in file_info
        assert "%relativeLocation%" in file_info
        if file_info.get("%fileUUID%") != "None":
            continue
        assert file_info.get("%relativeLocation%") == str(new_file)
        file_path = "".join(
            [
                package.replacement_path_string,
                "/objects",
                file_info.get("%relativeLocation%").split("objects")[1],
            ]
        )
        kwargs = {
            "uuid": uuid.uuid4(),
            "currentlocation": file_path,
            "filegrpuse": "original",
            "transfer_id": package.subid,
        }
        models.File.objects.create(**kwargs)
    assert (
        file_count == 2
    ), "Database and file objects were not returned by the generator"
    assert models.File.objects.filter(transfer_id=str(package.subid)).count() == 2

    # Simulate an additional file object being added later on in the transfer
    # in a sub directory of the objects folder, e.g. transcribe contents.
    sub_dir = objects_path / "subdir"
    sub_dir.mkdir()
    new_file = sub_dir / "another_new_file.txt"
    new_file.touch()
    for file_count, file_info in enumerate(package.files("/objects"), 1):
        file_path = "".join(
            [
                package.replacement_path_string,
                "/objects",
                file_info.get("%relativeLocation%").split("objects")[1],
            ]
        )
        if file_info.get("%fileUUID%") != "None":
            continue
        kwargs = {
            "uuid": uuid.uuid4(),
            "currentlocation": file_path,
            "filegrpuse": "original",
            "transfer_id": package.subid,
        }
        models.File.objects.create(**kwargs)
    assert (
        file_count == 3
    ), "Database and file objects were not returned by the generator"
    assert models.File.objects.filter(transfer_id=str(package.subid)).count() == 3

    # Now the database is updated, we will still have the same file count, but
    # all objects will be returned from the database and we will have uuids.
    for file_count, file_info in enumerate(package.files("/objects"), 1):
        if file_info.get("%fileUUID%") == "None":
            assert (
                False
            ), "Non-database entries returned from package.files(): {}".format(
                file_info
            )
    assert file_count == 3

    # Finally, let's just see if the scan works for a slightly larger no.
    # files, i.e. a number with an increment slightly larger than one.
    files = ["f1", "f2", "f3", "f4", "f5"]
    for file_ in files:
        new_file = objects_path / file_
        new_file.touch()
    new_count = 0
    for file_count, file_info in enumerate(package.files("/objects"), 1):
        if file_info.get("%fileUUID%") == "None":
            new_count += 1
    assert new_count == 5
    assert file_count == 8

    # Clean up state and ensure test doesn't interfere with other transfers
    # expected to be in the database, e.g. in test_queues.py.
    models.File.objects.filter(transfer_id=str(package.subid)).delete()


@pytest.mark.django_db(transaction=True)
def test_package_files_with_non_ascii_names(tmp_path, package):
    path = tmp_path / f"test-transfer-{package.subid}"
    package.current_path = str(path)

    # Add a file to the transfer with non-ascii name
    path.mkdir()
    objects = path / "objects"
    objects.mkdir()
    file_ = objects / "montréal.txt"
    file_.touch()

    # Create a File model instance for the transfer file
    kwargs = {
        "uuid": uuid.uuid4(),
        "currentlocation": "".join(
            [package.replacement_path_string, "/objects/montréal.txt"]
        ),
        "filegrpuse": "original",
        "transfer_id": package.subid,
    }
    models.File.objects.create(**kwargs)

    # Assert only one file is returned
    result = list(package.files("/objects"))
    assert len(result) == 1

    # And it is the file we just created
    assert result[0]["%fileUUID%"] == str(kwargs["uuid"])
    assert result[0]["%currentLocation%"] == kwargs["currentlocation"]
    assert result[0]["%fileGrpUse%"] == kwargs["filegrpuse"]



import pytest

from a3m.server.jobs import Job
from a3m.server.tasks import get_task_backend
from a3m.server.tasks import Task
from a3m.server.tasks import TaskBackend


class MockJob(Job):
    def __init__(self, *args, **kwargs):
        self.name = kwargs.pop("name", "")
        super().__init__(*args, **kwargs)

    def run(self, *args, **kwargs):
        pass


@pytest.fixture
def simple_job(request, mocker):
    return MockJob(mocker.Mock(), mocker.Mock(), mocker.Mock(), name="test_v0.0")


@pytest.fixture
def simple_task(request):
    return Task(
        "command",
        "a argument string",
        "/dev/stdoutfile",
        "/tmp/stderrfile",
        {r"%relativeLocation%": "testfile"},
        wants_output=False,
    )


def format_result(task_results):
    """Accepts task results as a tuple of (uuid, result_dict)."""
    response = {"task_results": {}}
    for task_uuid, task_data in task_results:
        task_uuid = str(task_uuid)
        response["task_results"][task_uuid] = task_data

    return response


# test_gearman_task_submission
# test_gearman_task_result_success
# test_gearman_task_result_error


def test_multiple_batches(simple_job, simple_task, mocker):
    mocker.patch("a3m.server.tasks.backends.pool_backend.Task.bulk_log")
    mocker.patch("a3m.server.tasks.backends.pool_backend.Task.write_output")
    mocker.patch("a3m.server.tasks.backends.pool_backend.init_counter_labels")
    mocker.patch.object(TaskBackend, "TASK_BATCH_SIZE", 2)

    def execute_command(task_name: str, batch_payload):
        assert task_name == "test_v0.0"
        return {
            "task_results": {
                task_id: {
                    "exitCode": 0,
                    "stdout": "stdout example",
                    "stderr": "stderr example",
                }
                for task_id, task in batch_payload["tasks"].items()
            }
        }

    execute_command = mocker.patch(
        "a3m.server.tasks.backends.pool_backend.execute_command",
        side_effect=execute_command,
    )

    backend = get_task_backend()

    for item in range(3):
        backend.submit_task(simple_job, simple_task)

    results = list(backend.wait_for_results(simple_job))
    assert execute_command.call_count == 2
    assert len(results) == 3
    assert results[0].done is True
    assert results[0].exit_code == 0
    assert results[1].done is True
    assert results[1].exit_code == 0
    assert results[2].done is True
    assert results[2].exit_code == 0



import pytest

from a3m.client.clientScripts.validate_file import main
from a3m.client.job import Job
from a3m.main.models import Event
from a3m.main.models import File
from a3m.main.models import FileFormatVersion
from a3m.main.models import SIP


@pytest.fixture
def sip(tmp_path):
    sip_dir = tmp_path / "sip"
    sip_dir.mkdir()
    # Create logs directory in the SIP.
    (sip_dir / "logs").mkdir()

    return SIP.objects.create(currentpath=str(sip_dir))


@pytest.fixture
def file_obj(tmp_path, sip):
    d = tmp_path / "dir"
    d.mkdir()
    txt_file = d / "file.txt"
    txt_file.write_text("hello world")

    f = File.objects.create(
        sip=sip, originallocation=txt_file, currentlocation=txt_file
    )
    f.fileformatversion_set.create(
        # Known format version with validation rule.
        format_version_id="45928c95-1fea-4b2b-af54-9aa3807e26a2",
    )

    return f


@pytest.fixture
def file_format_version(file_obj, format_version):
    FileFormatVersion.objects.create(file_uuid=file_obj, format_version=format_version)


@pytest.mark.django_db
def test_main(mocker, sip, file_obj):
    exit_status = 0
    stdout = '{"eventOutcomeInformation": "pass", "eventOutcomeDetailNote": "a note"}'
    stderr = ""
    execute_or_run = mocker.patch(
        "a3m.client.clientScripts.validate_file.executeOrRun",
        return_value=(exit_status, stdout, stderr),
    )
    job = mocker.Mock(spec=Job)
    file_type = "original"

    main(
        job=job,
        file_path=file_obj.currentlocation,
        file_uuid=file_obj.uuid,
        sip_uuid=sip.uuid,
        shared_path=sip.currentpath,
        file_type=file_type,
    )

    # Check the executed script.
    called_args = execute_or_run.call_args
    assert called_args.kwargs["type"] == "pythonScript"
    assert called_args.kwargs["printing"] is False
    assert called_args.kwargs["arguments"] == [file_obj.currentlocation]

    # Verify a PREMIS validation event was created with the output of the
    # validation command.
    assert (
        Event.objects.filter(
            file_uuid=file_obj.uuid,
            event_type="validation",
            event_outcome="pass",
            event_outcome_detail="a note",
        ).count()
        == 1
    )



from pathlib import Path
from uuid import uuid4

import pytest

from a3m.client.clientScripts import has_packages
from a3m.client.job import Job
from a3m.main.models import File
from a3m.main.models import Transfer


@pytest.fixture
def transfer(db, tmp_path):
    transfer_dir = tmp_path / "transfer"
    transfer_dir.mkdir()

    return Transfer.objects.create(currentlocation=str(transfer_dir))


@pytest.fixture
def compressed_file(db, transfer):
    # Simulate a compressed file being extracted to a directory with the same name.
    d = Path(transfer.currentlocation) / "compressed.zip"
    d.mkdir()

    # Place an extracted file in it.
    f = d / "file.txt"
    f.touch()

    # Create File models for the compressed and extracted files.
    result = File.objects.create(
        uuid=uuid4(), transfer=transfer, originallocation=d, currentlocation=d
    )
    File.objects.create(
        uuid=uuid4(), transfer=transfer, originallocation=f, currentlocation=f
    )

    # Create a file format version for the compressed file.
    # ce097bf8 is a fpr.formatversion for 7zip with an extraction rule.
    result.fileformatversion_set.create(
        format_version_id="ce097bf8-dc4d-4083-932e-82224890f26a"
    )

    return result


def test_main_detects_file_is_extractable_via_fpr(
    db, mocker, transfer, compressed_file
):
    job = mocker.Mock(spec=Job)

    result = has_packages.main(job, str(transfer.uuid))

    assert result == 0


def test_main_detects_file_was_already_extracted_from_unpacking_event(
    db, mocker, transfer, compressed_file
):
    job = mocker.Mock(spec=Job)

    extracted_file = File.objects.get(
        currentlocation__startswith=compressed_file.currentlocation,
        currentlocation__endswith="file.txt",
    )
    extracted_file.event_set.create(
        event_type="unpacking",
        event_detail=f"Unpacked from: {extracted_file.currentlocation} ({compressed_file.uuid})",
    )

    result = has_packages.main(job, str(transfer.uuid))

    assert result == 1



import concurrent.futures
import os
import threading
import uuid

import pytest
from django.utils import timezone

from a3m.api.transferservice.v1beta1.request_response_pb2 import ProcessingConfig
from a3m.main import models
from a3m.server.jobs import DirectoryClientScriptJob
from a3m.server.jobs import FilesClientScriptJob
from a3m.server.jobs import JobChain
from a3m.server.jobs import NextLinkDecisionJob
from a3m.server.packages import Package
from a3m.server.queues import PackageQueue
from a3m.server.tasks import TaskBackend
from a3m.server.workflow import load as load_workflow


FIXTURES_DIR = os.path.join(os.path.dirname(__file__), "fixtures")
INTEGRATION_TEST_PATH = os.path.join(FIXTURES_DIR, "workflow-integration-test.json")


class EchoBackend(TaskBackend):
    def __init__(self):
        self.tasks = {}

    def submit_task(self, job, task):
        if job.uuid not in self.tasks:
            self.tasks[job.uuid] = []
        self.tasks[job.uuid].append(task)

    def wait_for_results(self, job):
        for task in self.tasks[job.uuid]:
            task.exit_code = 0
            task.stdout = task.arguments
            task.stderr = task.arguments
            task.finished_timestamp = timezone.now()

            yield task


@pytest.fixture
def workflow(request):
    with open(INTEGRATION_TEST_PATH) as workflow_file:
        return load_workflow(workflow_file)


@pytest.fixture
def package_queue(request):
    executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
    return PackageQueue(executor, threading.Event(), debug=True)


class FakeUnit:
    def __init__(self, pk):
        self.pk = pk


@pytest.fixture
def package(request):
    return Package(
        "package-1",
        "file:///tmp/foobar-1.gz",
        ProcessingConfig(),
        models.Transfer.objects.create(pk=uuid.uuid4()),
        models.SIP.objects.create(pk=uuid.uuid4()),
    )


@pytest.fixture
def dummy_file_replacements(request):
    files = []
    for x in range(3):
        files.append(
            {
                r"%relativeLocation%": f"transfer_path/file{x}",
                r"%fileUUID%": str(uuid.uuid4()),
            }
        )

    return files


@pytest.mark.django_db(transaction=True)
def test_workflow_integration(
    mocker,
    settings,
    tmp_path,
    workflow,
    package_queue,
    package,
    dummy_file_replacements,
):
    # Setup our many mocks
    echo_backend = EchoBackend()
    settings.SHARED_DIRECTORY = str(tmp_path)
    settings.PROCESSING_DIRECTORY = str(tmp_path / "processing")
    mocker.patch.dict(
        "a3m.server.packages.BASE_REPLACEMENTS",
        {r"%processingDirectory%": settings.PROCESSING_DIRECTORY},
    )

    mock_get_task_backend = mocker.patch(
        "a3m.server.jobs.client.get_task_backend", return_value=echo_backend
    )
    mocker.patch.object(package, "files", return_value=dummy_file_replacements)

    # Schedule the first job
    initiator_link = workflow.get_initiator()
    first_job_chain = JobChain(package, workflow, initiator_link)
    job = next(first_job_chain)
    package_queue.schedule_job(job)

    assert package_queue.job_queue.qsize() == 1
    assert len(package_queue.active_packages) == 1
    assert package.uuid in package_queue.active_packages

    # Process the first job (DirectoryClientScriptJob)
    future = package_queue.process_one_job(timeout=1.0)
    concurrent.futures.wait([future], timeout=1.0)

    mock_get_task_backend.assert_called_once()
    task = echo_backend.tasks[job.uuid][0]

    assert isinstance(job, DirectoryClientScriptJob)
    assert job.exit_code == 0
    assert task.arguments == '"{}" "{}"'.format(
        settings.PROCESSING_DIRECTORY, package.subid
    )

    # Next job in chain should be queued
    assert package_queue.job_queue.qsize() == 1
    job = future.result()

    # Process the second job (FilesClientScriptJob)
    future = package_queue.process_one_job(timeout=1.0)
    concurrent.futures.wait([future], timeout=1.0)

    tasks = echo_backend.tasks[job.uuid]

    assert isinstance(job, FilesClientScriptJob)
    assert job.exit_code == 0
    assert len(tasks) == len(dummy_file_replacements)
    for task, replacement in zip(tasks, dummy_file_replacements):
        assert task.arguments == '"{}"'.format(replacement[r"%fileUUID%"])

    # Next job in chain should be queued
    assert package_queue.job_queue.qsize() == 1
    job = future.result()

    # Process the third job (DirectoryClientScriptJob)
    future = package_queue.process_one_job(timeout=1.0)
    concurrent.futures.wait([future], timeout=1.0)

    assert isinstance(job, DirectoryClientScriptJob)
    assert job.exit_code == 0

    # Next job in chain should be queued
    assert package_queue.job_queue.qsize() == 1
    job = future.result()

    # Process the fourth job (DirectoryClientScriptJob)
    future = package_queue.process_one_job(timeout=1.0)
    concurrent.futures.wait([future], timeout=1.0)

    assert isinstance(job, DirectoryClientScriptJob)
    assert job.exit_code == 0

    # Next job in chain should be queued
    assert package_queue.job_queue.qsize() == 1
    job = future.result()

    # Process the fifth job (NextLinkDecisionJob)
    future = package_queue.process_one_job(timeout=1.0)
    concurrent.futures.wait([future], timeout=1.0)

    assert isinstance(job, NextLinkDecisionJob)
    assert job.exit_code == 0

    # Next job in chain should be queued
    assert package_queue.job_queue.qsize() == 1
    job = future.result()

    # Process the sixth job (UpdateContextDecisionJob)
    future = package_queue.process_one_job(timeout=1.0)
    concurrent.futures.wait([future], timeout=1.0)

    # Out job chain should have been redirected to the final link
    assert job.job_chain.current_link.id == "f8e4c1ee-3e43-4caa-a664-f6b6bd8f156e"
    assert package_queue.job_queue.qsize() == 1
    job = future.result()

    assert isinstance(job, DirectoryClientScriptJob)
    assert job.exit_code is None

    # Process the last job (DirectoryClientScriptJob)
    future = package_queue.process_one_job(timeout=1.0)
    concurrent.futures.wait([future], timeout=1.0)

    assert job.exit_code == 0

    # Workflow is over; we're done
    assert package_queue.job_queue.qsize() == 0



#!/usr/bin/env python
"""HTTP gateway for a3m.

When executed, this module starts a simple HTTP server that submits transfers
to an embedded a3m server instance on every GET request received.

Usage::

    $ pip install a3m
    $ ./webapp.py
    $ curl 127.0.0.1:33892

"""
from http.server import BaseHTTPRequestHandler
from http.server import ThreadingHTTPServer
from threading import Thread

import grpc

from a3m.cli.common import init_django

init_django()  # This will not be needed in the future.
from a3m.server.runner import create_server
from a3m.server.rpc.client import Client


class RequestHandler(BaseHTTPRequestHandler):
    def __init__(self, *args):
        super().__init__(*args)

    def do_GET(self):
        self.send_response(200)
        self.send_header("Content-Type", "text/html")
        self.end_headers()
        try:
            resp = a3mc.submit(
                name="Test",
                url="https://github.com/artefactual/archivematica-sampledata/raw/master/SampleTransfers/ZippedDirectoryTransfers/DemoTransferCSV.zip",
            )
        except Exception as err:
            self.wfile.write(f"Error: {err}".encode())
        else:
            self.wfile.write(f"Transfer submitted! {resp.id}".encode())


a3md = create_server(
    bind_address="127.0.0.1:0",
    server_credentials=grpc.local_server_credentials(
        grpc.LocalConnectionType.LOCAL_TCP
    ),
    max_concurrent_packages=1,
    batch_size=125,
    queue_workers=3,
    grpc_workers=3,
)
a3md_thread = Thread(target=a3md.start)
a3md_thread.start()

httpd = ThreadingHTTPServer(("127.0.0.1", 0), RequestHandler)
httpd_thread = Thread(target=httpd.serve_forever)
httpd_thread.start()
print(f"Web server listening on port {httpd.server_port}/tcp.")

a3mc = Client(grpc.insecure_channel(f"127.0.0.1:{a3md.grpc_port}"))

a3md_thread.join()
